# -*- coding: utf-8 -*-

import asyncio
import aiohttp


def async_req(list_request, headers=None, auth=None, decode=False, schedule=0, request_type='GET'):

    async def _manage(_list, _auth, _sched):
        _tasks = []

        #Single request
        async def _single_request(_request,_sched):
            await asyncio.sleep(_sched)
            async with session.request(request_type, _request) as resp:
                assert resp.status == 200
                return await resp.read()

        #Creating all asyncio tasks
        if(_auth != None):
            _b_auth = aiohttp.BasicAuth(_auth['login'], password=_auth['password'])
        else:
            _b_auth = None
        async with aiohttp.ClientSession(loop=asyncio.get_event_loop(), auth=_b_auth, headers=headers) as session:
            _delay = 0
            for _req in _list:
                task = asyncio.create_task(_single_request(_req, _delay))
                _tasks.append(task)
                _delay = _delay + _sched
            await asyncio.gather(*_tasks)
        return _tasks

    _completed_tasks = asyncio.run(_manage(list_request, auth, schedule))
    if decode:
        _tasks_to_return = [task.result().decode() for task in _completed_tasks]
    else:
        _tasks_to_return = [task.result() for task in _completed_tasks]
    return _tasks_to_return



#!/usr/bin/env python
# coding: utf-8

import NGY.Map_func
import NGY.Magic_Request
import NGY.Tiles
from urllib.parse import urlencode
import sqlite3
import os


class Request:

    def __init__(self, bounds, zoom_list, layer="GEOGRAPHICALGRIDSYSTEMS.MAPS.SCAN25TOUR.CV",
                 style="normal", projection="PM", format="jpeg"):
        self.bounds = (min(bounds[1], bounds[3]),
                       min(bounds[0], bounds[2]),
                       max(bounds[1], bounds[3]),
                       max(bounds[0], bounds[2]))  # WGS 84 Coordinates
        self.zoom_list = zoom_list
        self.layer = layer
        self.style = style
        self.projection = projection
        self.format = format
        self._tile_set = None
        self._request_list = []
        self._base_url = "https://wxs.ign.fr/an7nvfzojv5wa96dsga5nk8w/geoportail/wmts"

        self._headers = {
            'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
            'Accept-Encoding': 'gzip, deflate',
            'Accept-Language': 'en-US,en;q=0.5',
            'Connection': 'keep-alive',
            'Host': 'wxs.ign.fr',
            'Referer': 'http://www.geoportail.gouv.fr/swf/geoportal-visu-1.3.2.swf',
            'User-Agent': 'Mozilla/5.0 (Windows NT 6.1; WOW64; rv:22.0) Gecko/20100101 Firefox/22.0',
        }

    def get_number_of_tiles(self):
        self._build_request_list()
        return len(self._request_list)

    def _build_request_list(self):
        self._request_list = []
        _default_value = {
            'layer': self.layer,
            'style': self.style,
            'tilematrixset': self.projection,
            'Service': 'WMTS',
            'Request': 'GetTile',
            'Version': '1.0.0',
            'Format': 'image/' + self.format
        }

        for k in self.zoom_list:
            top_left_corner = NGY.Map_func.Coord._make(
                NGY.Map_func.convert_coord((min(self.bounds[0], self.bounds[2]), max(self.bounds[1], self.bounds[3])),
                                           NGY.Map_func.MATRIX_SET_DATA, k, self.projection))

            bot_right_corner = NGY.Map_func.Coord._make(
                NGY.Map_func.convert_coord((max(self.bounds[0], self.bounds[2]), min(self.bounds[1], self.bounds[3])),
                                           NGY.Map_func.MATRIX_SET_DATA, k, self.projection))

            def url_build(i, j, zoom):
                _values = _default_value
                _values['TileCol'] = i
                _values['TileRow'] = j
                _values['TileMatrix'] = zoom
                return self._base_url + '?' + urlencode(_values)

            for j in range(top_left_corner.lat, bot_right_corner.lat + 1):
                for i in range(top_left_corner.lon, bot_right_corner.lon + 1):
                    self._request_list.append({'x': i, 'y': j, 'z': k, 'url': url_build(i, j, k)})

    def fetch_data(self):
        if not self._request_list:
            self._build_request_list()
        _data = NGY.Magic_Request.async_req([_item['url'] for _item in self._request_list], headers=self._headers)
        # TODO : To optimize
        _temp = list(zip(self._request_list, _data))
        _temp2 = [NGY.Tiles.Tile(_item[0]['x'], _item[0]['y'], _item[0]['z'], _item[1], format=self.format) for _item in _temp]
        self._tile_set = NGY.Tiles.Tileset(_temp2)

    def save(self, file, name, meta=None ):
        # See https://github.com/mapbox/mbtiles-spec/blob/master/1.3/spec.md
        _format_trans = dict(jpeg="jpg", png="png", pbf="pbf", webp="webp")
        _meta = dict(name=name,
                     format=_format_trans[self.format],
                     bounds=", ".join(str(v) for v in self.bounds),
                     minzoom=min(self.zoom_list),
                     maxzoom=max(self.zoom_list),
                     center=", ".join(str(v) for v in [round((self.bounds[0] + self.bounds[2])/2, 7),
                             round((self.bounds[1] + self.bounds[3])/2, 7),
                             max(self.zoom_list)])
                     )
        if meta is not None:
            _meta = {**_meta, **meta}

        if os.path.isfile(file):
            os.remove(file)
        conn = sqlite3.connect(file)
        c = conn.cursor()
        # Create db structure
        # Create table metadata (need name, format, bounds)
        c.execute('CREATE TABLE metadata (name text, value text)')
        # Create table tiles
        c.execute('CREATE TABLE tiles (zoom_level integer, tile_column integer, tile_row integer, tile_data blob)')
        c.execute('CREATE UNIQUE INDEX tile_index on tiles (zoom_level, tile_column, tile_row)')
        for key, val in _meta.items():
            c.execute('INSERT INTO metadata (name,value) VALUES (?,?)', (key, val))
        for _tile in self._tile_set.get_tiles_list():
            c.execute('INSERT INTO tiles (tile_column , tile_row , zoom_level , tile_data) VALUES (?,?,?,?)',
                      (*_tile.get_TMS_coordinates(), sqlite3.Binary(_tile.get_data())))

        # Save (commit) the changes
        conn.commit()

        # We can also close the connection if we are done with it.
        # Just be sure any changes have been committed or they will be lost.
        conn.close()


#!/usr/bin/env python
# coding: utf-8

import NGY

my_request = NGY.Request(( 43.553859, 1.364688, 43.394638, 1.580797),
                         zoom_list=range(6,16),
                         layer="GEOGRAPHICALGRIDSYSTEMS.MAPS.SCAN25TOUR.CV",
                         style="normal",
                         projection="PM",
                         format="jpeg")

my_request.fetch_data()
my_request.save('Lacroix-Falgarde.mbtiles', 'Lacroix-Falgarde - 02/06/2023')




#!/usr/bin/env python
# coding: utf-8

import os
from collections import namedtuple
from lxml import etree
import pyproj

# XML DATA
WMTS_FILE_PATH = os.path.join(os.path.abspath(os.path.dirname(__file__)),"data", "wmts.xml")
_DEF_NS = {"foo": "http://www.opengis.net/wmts/1.0", "ows": "http://www.opengis.net/ows/1.1"}

# Type def
Coord = namedtuple('Coord', ['lon', 'lat'])

# to get more info : https://wxs.ign.fr/an7nvfzojv5wa96dsga5nk8w/geoportail/wmts?Service=WMTS&Request=GetCapabilities
# &Version=1.0.0&Layer=GEOGRAPHICALGRIDSYSTEMS.MAPS.SCAN25TOUR.CV


def get_scale_denom_info(file):
    """Return dict with tile matrix set identifier, zoom(scale), scale denom(meters/pixel)"""
    data = {}
    tree = etree.parse(file)

    tile_matrix_sets = tree.xpath('/foo:Capabilities/foo:Contents/foo:TileMatrixSet', namespaces=_DEF_NS)
    for tilematrixset in tile_matrix_sets:
        identifier = tilematrixset.find('ows:Identifier', namespaces=_DEF_NS).text
        supportedcrs = tilematrixset.find('ows:SupportedCRS', namespaces=_DEF_NS).text
        data[identifier] = {}
        data[identifier]['supportedcrs'] = supportedcrs
        TileMatrixes = tilematrixset.findall('foo:TileMatrix', namespaces=_DEF_NS)
        for tilematrix in TileMatrixes:
            scale = tilematrix.find('ows:Identifier', namespaces=_DEF_NS).text
            # Dénominateur d'échelle = résolution / taille pixel ; taille de pixel arbitraire = 0.00028 m
            scale_denom = float(tilematrix.find('foo:ScaleDenominator', namespaces=_DEF_NS).text) * 0.00028
            data[identifier][scale] = scale_denom
            str_ref = tilematrix.find('foo:TopLeftCorner', namespaces=_DEF_NS).text
            (data[identifier]['X_ref'], data[identifier]['Y_ref']) = map(float, str_ref.split(" "))
    return data

# Get all information about tile sets
MATRIX_SET_DATA = get_scale_denom_info(WMTS_FILE_PATH)

def convert_coord(long_lat, matrix_data, zoom, projection):
    """convert decimals coords into Web Mercator coords into tile coords"""
    # See https://geoservices.ign.fr/documentation/geoservices/wmts.html

    long_lat_ref = pyproj.Proj(init='epsg:4326')

    final_proj = pyproj.Proj(init=matrix_data[projection]['supportedcrs'])
    tile_res = matrix_data[projection][str(zoom)] * 256  # 256 pixel per tile

    # Web Mercator
    (x, y) = pyproj.transform(long_lat_ref, final_proj, long_lat[0], long_lat[1])

    # Get the top left corner
    (x0, y0) = (matrix_data[projection]['X_ref'], matrix_data[projection]['Y_ref'])

    # Tile coords depending on matrix_data (see get_scale_denom_info function)
    (xf, yf) = (int((x - x0) / tile_res), int((y0 - y) / tile_res))
    # ("Zoom: " + str(zoom) + " Tile Rez: " + str(tile_res) + " X,Y: " + str(xf) + "," + str(yf))

    return (xf, yf)


#!/usr/bin/env python
# coding: utf-8


class Tile:

    def __init__(self, x, y, z, data, format="jpeg"):
        self.x = x
        self.y = y
        self.z = z
        self.data = data
        self.format = format

    def get_coordinates(self):
        return self.x, self.y, self.z

    def get_TMS_coordinates(self):
        new_y = pow(2, self.z) - 1 - self.y
        return self.x, new_y, self.z

    def get_data(self):
        return self.data


class Tileset:

    def __init__(self, tile_list):
        self._tile_list = tile_list
        for tile in self._tile_list:
            if not isinstance(tile, Tile):
                raise Exception("Tile is not an instance of Tile class")

    def get_tiles_list(self):
        return self._tile_list
